from . import runtime


class Statement:
    pass


class ExpressionStmt(Statement):

    def __init__(self, expr):
        self.expr = expr

    def exec(self, env):
        self.expr.eval(env)


class VarDecStmt(Statement):

    def __init__(self, name_token, initializer_expr):
        self.name_token = name_token
        self.initializer = initializer_expr

    def exec(self, env):
        value = None
        if self.initializer is not None:
            value = self.initializer.eval(env)
        env.define(self.name_token.text, value)


class BlockStmt(Statement):

    def __init__(self, statements):
        self.statements = statements

    def exec(self, env):
        block_env = runtime.Environment(env)
        for statement in self.statements:
            statement.exec(block_env)


class EchoStmt(Statement):

    def __init__(self, expr_list):
        self.expr_list = expr_list

    def exec(self, env):
        values = [expr.eval(env) for expr in self.expr_list]
        output = " ".join(runtime.stringify(value) for value in values)
        print(output)


class IfStmt(Statement):

    def __init__(self, condition_expr, then_stmt, else_stmt):
        self.condition = condition_expr
        self.then_stmt = then_stmt
        self.else_stmt = else_stmt

    def exec(self, env):
        if runtime.is_truthy(self.condition.eval(env)):
            self.then_stmt.exec(env)
        elif self.else_stmt is not None:
            self.else_stmt.exec(env)


class WhileStmt(Statement):

    def __init__(self, condition_expr, body_stmt):
        self.condition = condition_expr
        self.body_stmt = body_stmt

    def exec(self, env):
        try:
            while runtime.is_truthy(self.condition.eval(env)):
                try:
                    self.body_stmt.exec(env)
                except runtime.Continue:
                    pass
        except runtime.Break:
            pass


class ForStmt(Statement):

    def __init__(self, initializer_stmt, condition_expr, increment_expr, body_stmt):
        self.initializer = initializer_stmt
        self.condition = condition_expr
        self.increment = increment_expr
        self.body = body_stmt

    def exec(self, env):
        if self.initializer is not None:
            env = runtime.Environment(env)
            self.initializer.exec(env)
        try:
            while runtime.is_truthy(self.condition.eval(env)):
                try:
                    self.body.exec(env)
                except runtime.Continue:
                    pass
                if self.increment is not None:
                    self.increment.eval(env)
        except runtime.Break:
            pass


class ReturnStmt(Statement):

    def __init__(self, value_expr):
        self.value = value_expr

    def exec(self, env):
        if self.value is None:
            raise runtime.Return(None)
        else:
            raise runtime.Return(self.value.eval(env))


class BreakStmt(Statement):

    def exec(self, env):
        raise runtime.Break()


class ContinueStmt(Statement):

    def exec(self, env):
        raise runtime.Continue()


class FunctionStmt(Statement):

    def __init__(self, name_token, parameter_token_list, body_block_stmt):
        self.name_token = name_token
        self.parameters = parameter_token_list
        self.body = body_block_stmt

    def exec(self, env):
        func_obj = runtime.PyroFunction(self, env)
        env.define(self.name_token.text, func_obj)


class ClassStmt(Statement):

    def __init__(self, name_token, superclass_expr, method_stmt_list, field_dec_list):
        self.name_token = name_token
        self.superclass_expr = superclass_expr
        self.method_stmts = method_stmt_list
        self.field_decs = field_dec_list

    def exec(self, env):
        class_env = runtime.Environment(env)
        superclass_obj = None
        if self.superclass_expr is not None:
            superclass_obj = self.superclass_expr.eval(env)
            if not isinstance(superclass_obj, runtime.PyroClass):
                token = self.superclass_expr.token
                runtime.error(f"Invalid superclass name '{token.text}'.", self.token)
            class_env.define("super", superclass_obj)
        method_objs = {}
        for func_stmt in self.method_stmts:
            func_obj = runtime.PyroFunction(func_stmt, class_env)
            method_objs[func_stmt.name_token.text] = func_obj
        class_obj = runtime.PyroClass(self, superclass_obj, method_objs, class_env)
        env.define(self.name_token.text, class_obj)



import sys

from . import runtime
from . import lexer
from . import parser
from . import argslib


__version__ = "0.1.0"


def cmd_debug_tokens(cmd_name, cmd_parser):
    if cmd_parser.args:
        source = open(cmd_parser.args[0]).read()
    else:
        source = sys.stdin.read()
    for token in lexer.Lexer(source).tokenize():
        print(token)


def cmd_debug_expr(cmd_name, cmd_parser):
    if cmd_parser.args:
        source = open(cmd_parser.args[0]).read()
    else:
        source = sys.stdin.read()
    tokens = lexer.Lexer(source).tokenize()
    expression = parser.Parser(tokens).expression()
    print(expression)


def run(argparser):
    if argparser.args:
        source = open(argparser.args[0]).read()
    else:
        source = sys.stdin.read()
    tokens = lexer.Lexer(source).tokenize()
    statements = parser.Parser(tokens).parse()
    runtime.Interpreter().run(statements)


def main():
    argparser = argslib.ArgParser(version=__version__)
    argparser.command("debug_tokens", callback=cmd_debug_tokens)
    argparser.command("debug_expr", callback=cmd_debug_expr)
    argparser.parse()
    if not argparser.command_name:
        run(argparser)




##
# This module makes the package directly executable. To run a package located
# on Python's import search path use:
#
#   $ python -m proto
#
# To run an arbitrary `proto` package use:
#
#   $ python /path/to/proto/package
#
# Note that we need to manually add the package's parent directory to the
# search path to make the import work.
##

import os
import sys

sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))
import proto
sys.path.pop(0)

proto.main()




from . import runtime
import sys


class Expression:

    def __str__(self):
        return f"{self.__class__.__name__}"


class UnaryExpr(Expression):

    def __init__(self, operator_token, expr):
        self.op = operator_token
        self.expr = expr

    def __str__(self):
        return f"({self.op.text} {self.expr})"

    def eval(self, env):
        value = self.expr.eval(env)
        if self.op.type == "MINUS":
            runtime.check_operand_is_float(value, self.op)
            return -value
        elif self.op.type == "BANG":
            return not runtime.is_truthy(value)


class BinaryExpr(Expression):

    def __init__(self, left_expr, operator_token, right_expr):
        self.left = left_expr
        self.op = operator_token
        self.right = right_expr

    def __str__(self):
        return f"({self.op.text} {self.left} {self.right})"

    def eval(self, env):
        left = self.left.eval(env)
        right = self.right.eval(env)
        if self.op.type == "GREATER":
            runtime.check_operands_are_floats_or_strings(left, right, self.op)
            return left > right
        elif self.op.type == "GREATER_EQUAL":
            runtime.check_operands_are_floats_or_strings(left, right, self.op)
            return left >= right
        elif self.op.type == "LESS":
            runtime.check_operands_are_floats_or_strings(left, right, self.op)
            return left < right
        elif self.op.type == "LESS_EQUAL":
            runtime.check_operands_are_floats_or_strings(left, right, self.op)
            return left <= right
        elif self.op.type in ("PLUS", "PLUS_EQUAL"):
            runtime.check_operands_are_floats_or_strings(left, right, self.op)
            return left + right
        elif self.op.type in ("MINUS", "MINUS_EQUAL"):
            runtime.check_operands_are_floats(left, right, self.op)
            return left - right
        elif self.op.type == "SLASH":
            runtime.check_operands_are_floats(left, right, self.op)
            return left / right
        elif self.op.type == "STAR":
            runtime.check_operands_are_floats(left, right, self.op)
            return left * right
        elif self.op.type == "BANG_EQUAL":
            return left != right
        elif self.op.type == "EQUAL_EQUAL":
            return left == right


class LogicalExpr(Expression):

    def __init__(self, left_expr, operator_token, right_expr):
        self.left = left_expr
        self.op = operator_token
        self.right = right_expr

    def __str__(self):
        return f"({self.op.text} {self.left} {self.right})"

    def eval(self, env):
        left = self.left.eval(env)
        if self.op.type == "OR":
            if runtime.is_truthy(left):
                return left
        else:
            if not runtime.is_truthy(left):
                return left
        return self.right.eval(env)


class LiteralExpr(Expression):

    def __init__(self, token, value):
        self.token = token
        self.value = value

    def __str__(self):
        if self.token.type in ("NULL", "TRUE", "FALSE"):
            return self.token.type.lower()
        else:
            return repr(self.value)

    def eval(self, env):
        return self.value


class GroupingExpr(Expression):

    def __init__(self, expr):
        self.expr = expr

    def __str__(self):
        return f"(GROUP {self.expr})"

    def eval(self, env):
        return self.expr.eval(env)


class VariableExpr(Expression):

    def __init__(self, token):
        self.token = token

    def __str__(self):
        return self.token.text

    def eval(self, env):
        return env.get(self.token)


class AssignExpr(Expression):

    def __init__(self, token, value_expr):
        self.token = token
        self.value = value_expr

    def __str__(self):
        return f"(= {self.token.text} {self.value})"

    def eval(self, env):
        value = self.value.eval(env)
        env.set(self.token, value)
        return value


class CallExpr(Expression):

    def __init__(self, callee_expr, left_paren_token, arg_expr_list):
        self.callee = callee_expr
        self.token = left_paren_token
        self.args = arg_expr_list

    def __str__(self):
        return f"(CALL {self.callee})"

    def eval(self, env):
        callee = self.callee.eval(env)
        if not hasattr(callee, "call"):
            runtime.error(f"Object is not callable.", self.token)
        if len(self.args) != callee.arity():
            runtime.error(f"Invalid number of arguments.", self.token)
        args = []
        for arg in self.args:
            args.append(arg.eval(env))
        return callee.call(args)


class GetAttrExpr(Expression):

    def __init__(self, object_expr, name_token):
        self.object = object_expr
        self.name = name_token

    def __str__(self):
        return f"(GET {self.object} {self.name.text})"

    def eval(self, env):
        instance_obj = self.object.eval(env)
        if isinstance(instance_obj, runtime.PyroInstance):
            return instance_obj.get(self.name)
        runtime.error(f"Invalid attribute '{self.name.text}', not an instance object.", self.name)


class SetAttrExpr(Expression):

    def __init__(self, object_expr, name_token, value_expr):
        self.object = object_expr
        self.name = name_token
        self.value = value_expr

    def __str__(self):
        return f"(SET {self.object} {self.name.text} {self.value})"

    def eval(self, env):
        instance_obj = self.object.eval(env)
        if isinstance(instance_obj, runtime.PyroInstance):
            value = self.value.eval(env)
            instance_obj.set(self.name, value)
            return value
        runtime.error(f"Invalid attribute '{self.name.text}', not an instance object.", self.name)


class SelfExpr(Expression):

    def __init__(self, token):
        self.token = token

    def __str__(self):
        return "self"

    def eval(self, env):
        return env.get(self.token)


class SuperExpr(Expression):

    def __init__(self, super_token, attr_token):
        self.super_token = super_token
        self.attr_token = attr_token

    def eval(self, env):
        superclass_obj = env.get(self.super_token)
        instance_obj = env.get_str("self")
        method = superclass_obj.find_and_bind_method(instance_obj, self.attr_token.text)
        if method is None:
            runtime.error(f"Invalid superclass method name '{self.attr_token.text}'.", self.attr_token)
        return method


class ConditionalExpr(Expression):

    def __init__(self, condition_expr, true_branch_expr, false_branch_expr):
        self.condition = condition_expr
        self.true_branch = true_branch_expr
        self.false_branch = false_branch_expr

    def eval(self, env):
        if self.condition.eval(env):
            return self.true_branch.eval(env)
        else:
            return self.false_branch.eval(env)




import sys


keywords = {
    "and",
    "break",
    "class",
    "continue",
    "def",
    "echo",
    "else",
    "false",
    "for",
    "if",
    "null",
    "or",
    "return",
    "self",
    "super",
    "true",
    "var",
    "while",
}


def is_digit(char):
    return ord(char) >= ord('0') and ord(char) <= ord('9')


def is_alpha(char):
    if ord(char) >= ord('a') and ord(char) <= ord('z'):
        return True
    if ord(char) >= ord('A') and ord(char) <= ord('Z'):
        return True
    if char == '_' or char == '$':
        return True
    return False


def is_alphanumeric(char):
    return is_digit(char) or is_alpha(char)


class Token:

    def __init__(self, token_type, token_text, line_number):
        self.type = token_type
        self.text = token_text
        self.line = line_number

    def __str__(self):
        return f'{self.line} :: {self.type} :: {self.text}'


class Lexer:

    def __init__(self, source_text):
        self.source_text = source_text
        self.tokens = []
        self.start_index = 0
        self.current_index = 0
        self.line_number = 1

    def tokenize(self):
        while not self.is_at_end():
            self.start_index = self.current_index
            self.read_next_token()
        self.tokens.append(Token("EOF", "", self.line_number))
        return self.tokens

    def is_at_end(self):
        return self.current_index >= len(self.source_text)

    def add_token(self, token_type):
        token_text = self.source_text[self.start_index:self.current_index]
        self.tokens.append(Token(token_type, token_text, self.line_number))

    def read_next_token(self):
        char = self.advance()

        # Single character tokens.
        if char == "(":
            self.add_token("LEFT_PAREN")
        elif char == ")":
            self.add_token("RIGHT_PAREN")
        elif char == "{":
            self.add_token("LEFT_BRACE")
        elif char == "}":
            self.add_token("RIGHT_BRACE")
        elif char == ",":
            self.add_token("COMMA")
        elif char == ".":
            self.add_token("DOT")
        elif char == ";":
            self.add_token("SEMICOLON")
        elif char == "*":
            self.add_token("STAR")
        elif char == "/":
            self.add_token("SLASH")
        elif char == "?":
            self.add_token("QUESTION")
        elif char == ":":
            self.add_token("COLON")

        # Single or double character tokens.
        elif char == "+":
            self.add_token("PLUS_EQUAL" if self.match("=") else "PLUS")
        elif char == "-":
            self.add_token("MINUS_EQUAL" if self.match("=") else "MINUS")
        elif char == "!":
            self.add_token("BANG_EQUAL" if self.match("=") else "BANG")
        elif char == "=":
            self.add_token("EQUAL_EQUAL" if self.match("=") else "EQUAL")
        elif char == "<":
            self.add_token("LESS_EQUAL" if self.match("=") else "LESS")
        elif char == ">":
            self.add_token("GREATER_EQUAL" if self.match("=") else "GREATER")

        # Discard comments.
        elif char == "#":
            while self.peek() != "\n" and not self.is_at_end():
                self.advance()

        # Discard whitespace.
        elif char == " " or char == "\r" or char == "\t" or char == "\n":
            pass

        # Strings.
        elif char == '"':
            self.read_string()

        # Numbers.
        elif is_digit(char):
            self.read_number()

        # Identifiers & keywords.
        elif is_alpha(char):
            self.read_identifier()

        else:
            sys.exit(f"Syntax Error: Line {self.line_number}. Unexpected character '{char}'.")

    def advance(self):
        next_char = self.source_text[self.current_index]
        if next_char == "\n":
            self.line_number += 1
        self.current_index += 1
        return next_char

    def match(self, char):
        if self.peek() == char:
            self.advance()
            return True
        return False

    def peek(self):
        if self.is_at_end():
            return None
        return self.source_text[self.current_index]

    def peek_next(self):
        if self.current_index + 1 >= len(self.source_text):
            return None
        return self.source_text[self.current_index + 1]

    def read_string(self):
        start_line = self.line_number
        while True:
            if self.is_at_end():
                sys.exit(f"Syntax Error: Unterminated string, opened in line {start_line}.")
            if self.advance() == '"':
                break
        self.add_token("STRING")

    def read_number(self):
        while is_digit(self.peek()):
            self.advance()
        if self.peek() == '.' and is_digit(self.peek_next()):
            self.advance()
            while is_digit(self.peek()):
                self.advance()
        self.add_token("NUMBER")

    def read_identifier(self):
        while is_alphanumeric(self.peek()):
            self.advance()
        lexeme = self.source_text[self.start_index:self.current_index]
        if lexeme in keywords:
            self.add_token(lexeme.upper())
        else:
            self.add_token("IDENTIFIER")



import sys
from .expressions import *
from .statements import *


# Recursive descent parser.
class Parser:

    def __init__(self, token_list):
        self.tokens = token_list
        self.current = 0

    def parse(self):
        statements = []
        while not self.is_at_end():
            statements.append(self.declaration())
        return statements

    # ------------------ #
    # Statement parsers. #
    # ------------------ #

    def declaration(self):
        if self.match("VAR"):
            return self.variable_declaration()
        if self.match("DEF"):
            return self.function_declaration("function")
        if self.match("CLASS"):
            return self.class_declaration()
        self.error("Expected a declaration.")

    def variable_declaration(self):
        name_token = self.consume("IDENTIFIER", "Expected a variable name after 'var'.")
        initializer_expr = None
        if self.match("EQUAL"):
            initializer_expr = self.expression()
        self.consume("SEMICOLON", "Expected ';' after a variable declaration.")
        return VarDecStmt(name_token, initializer_expr)

    def function_declaration(self, kind="function"):
        name_token = self.consume("IDENTIFIER", f"Expected {kind} name.")
        self.consume("LEFT_PAREN", f"Expected '(' after {kind} name.")
        parameters = []
        if not self.check("RIGHT_PAREN"):
            while True:
                parameters.append(self.consume("IDENTIFIER", "Expected parameter name."))
                if not self.match("COMMA"):
                    break
        self.consume("RIGHT_PAREN", "Expected ')' after parameters.")
        self.consume("LEFT_BRACE", f"Expected '{{' before {kind} body.")
        body = self.block()
        return FunctionStmt(name_token, parameters, body)

    def class_declaration(self):
        name_token = self.consume("IDENTIFIER", "Expected class name.")
        superclass_expr = None
        if self.match("LESS"):
            self.consume("IDENTIFIER", "Expected superclass name.")
            superclass_expr = VariableExpr(self.previous())
        self.consume("LEFT_BRACE", "Expected '{' before class body.")
        methods = []
        fields = []
        while not self.check("RIGHT_BRACE") and not self.is_at_end():
            if self.match("VAR"):
                fields.append(self.variable_declaration())
            elif self.match("DEF"):
                methods.append(self.function_declaration("method"))
            else:
                self.error("Expected a method or variable declaration.")
        self.consume("RIGHT_BRACE", "Expected '}' after class body.")
        return ClassStmt(name_token, superclass_expr, methods, fields)

    def block(self):
        statements = []
        while not self.check("RIGHT_BRACE") and not self.is_at_end():
            statements.append(self.body_statement())
        self.consume("RIGHT_BRACE", "Expected '}' after a block.")
        return BlockStmt(statements)

    def body_statement(self):
        if self.match("LEFT_BRACE"):
            return self.block()
        elif self.match("VAR"):
            return self.variable_declaration()
        elif self.match("ECHO"):
            return self.echo_statement()
        elif self.match("IF"):
            return self.if_statement()
        elif self.match("WHILE"):
            return self.while_statement()
        elif self.match("FOR"):
            return self.for_statement()
        elif self.match("BREAK"):
            return self.break_statement()
        elif self.match("CONTINUE"):
            return self.continue_statement()
        elif self.match("RETURN"):
            return self.return_statement()
        return self.expression_statement()

    def echo_statement(self):
        expressions = []
        if not self.check("SEMICOLON"):
            while True:
                expressions.append(self.expression())
                if not self.match("COMMA"):
                    break
        self.consume("SEMICOLON", "Expected ';' after 'echo' statement.")
        return EchoStmt(expressions)

    def if_statement(self):
        condition_expr = self.expression()
        self.consume("LEFT_BRACE", "Expected block after 'if' condition.")
        then_branch = self.block()
        else_branch = None
        if self.match("ELSE"):
            if self.match("IF"):
                else_branch = self.if_statement()
            else:
                self.consume("LEFT_BRACE", "Expected block after 'else'.")
                else_branch = self.block()
        return IfStmt(condition_expr, then_branch, else_branch)

    def while_statement(self):
        condition_expr = self.expression()
        self.consume("LEFT_BRACE", "Expected block after 'while'.")
        body = self.block()
        return WhileStmt(condition_expr, body)

    def for_statement(self):
        if self.match("SEMICOLON"):
            initializer_stmt = None
        elif self.match("VAR"):
            initializer_stmt = self.variable_declaration()
        else:
            initializer_stmt = self.expression_statement()

        if self.check("SEMICOLON"):
            condition_expr = LiteralExpr(None, True)
        else:
            condition_expr = self.expression()
        self.consume("SEMICOLON", "Expected ';' after loop condition.")

        if self.check("LEFT_BRACE"):
            increment_expr = None
        else:
            increment_expr = self.expression()

        self.consume("LEFT_BRACE", "Expected block after 'for'.")
        body = self.block()
        return ForStmt(initializer_stmt, condition_expr, increment_expr, body)

    def break_statement(self):
        self.consume("SEMICOLON", "Expected ';' after 'break'.")
        return BreakStmt()

    def continue_statement(self):
        self.consume("SEMICOLON", "Expected ';' after 'continue'.")
        return ContinueStmt()

    def return_statement(self):
        if self.match("SEMICOLON"):
            return ReturnStmt(None)
        value_expr = self.expression()
        self.consume("SEMICOLON", "Expected ';' after return value.")
        return ReturnStmt(value_expr)

    def expression_statement(self):
        expr = self.expression()
        self.consume("SEMICOLON", "Expected ';' after expression.")
        return ExpressionStmt(expr)

    # ------------------- #
    # Expression parsers. #
    # ------------------- #

    def expression(self):
        return self.assignment()

    def assignment(self):
        expr = self.conditional()
        if self.match("EQUAL", "PLUS_EQUAL", "MINUS_EQUAL"):
            operator_token = self.previous()
            value_expr = self.assignment()
            if operator_token.type in ("PLUS_EQUAL", "MINUS_EQUAL"):
                value_expr = BinaryExpr(expr, operator_token, value_expr)
            if isinstance(expr, VariableExpr):
                return AssignExpr(expr.token, value_expr)
            elif isinstance(expr, GetAttrExpr):
                return SetAttrExpr(expr.object, expr.name, value_expr)
            else:
                self.error("Invalid assignment target.", operator_token)
        return expr

    def conditional(self):
        expr = self.logical()
        if self.match("QUESTION"):
            true_branch = self.logical()
            self.consume("COLON", "Expected ':' after '?'.")
            false_branch = self.logical()
            return ConditionalExpr(expr, true_branch, false_branch)
        return expr

    def logical(self):
        expr = self.equality()
        while self.match("AND", "OR"):
            operator_token = self.previous()
            right_expr = self.equality()
            expr = LogicalExpr(expr, operator_token, right_expr)
        return expr

    def equality(self):
        expr = self.comparative()
        while self.match("BANG_EQUAL", "EQUAL_EQUAL"):
            operator_token = self.previous()
            right_expr = self.comparative()
            expr = BinaryExpr(expr, operator_token, right_expr)
        return expr

    def comparative(self):
        expr = self.additive()
        while self.match("GREATER", "GREATER_EQUAL", "LESS", "LESS_EQUAL"):
            operator_token = self.previous()
            right = self.additive()
            expr = BinaryExpr(expr, operator_token, right)
        return expr

    def additive(self):
        expr = self.multiplicative()
        while self.match("MINUS", "PLUS"):
            operator_token = self.previous()
            right = self.multiplicative()
            expr = BinaryExpr(expr, operator_token, right)
        return expr

    def multiplicative(self):
        expr = self.unary()
        while self.match("SLASH", "STAR"):
            operator_token = self.previous()
            right = self.unary()
            expr = BinaryExpr(expr, operator_token, right)
        return expr

    def unary(self):
        if self.match("BANG", "MINUS"):
            operator_token = self.previous()
            expr = self.unary()
            return UnaryExpr(operator_token, expr)
        return self.call()

    def call(self):
        expr = self.primary()
        while True:
            if self.match("LEFT_PAREN"):
                expr = self.finish_call(expr)
            elif self.match("DOT"):
                name = self.consume("IDENTIFIER", "Expected property name after '.'.")
                expr = GetAttrExpr(expr, name)
            else:
                break
        return expr

    def finish_call(self, callee_expr):
        left_paren_token = self.previous()
        arguments = []
        if not self.check("RIGHT_PAREN"):
            while True:
                arguments.append(self.expression())
                if not self.match("COMMA"):
                    break
        self.consume("RIGHT_PAREN", "Expected ')' after arguments.")
        return CallExpr(callee_expr, left_paren_token, arguments)

    def primary(self):
        if self.match("SELF"):
            return SelfExpr(self.previous())
        if self.match("SUPER"):
            super_token = self.previous()
            self.consume("DOT", "Expected '.' after 'super'.")
            attr_token = self.consume("IDENTIFIER", "Expected superclass method name.")
            return SuperExpr(super_token, attr_token)
        if self.match("TRUE"):
            return LiteralExpr(self.previous(), True)
        if self.match("FALSE"):
            return LiteralExpr(self.previous(), False)
        if self.match("NULL"):
            return LiteralExpr(self.previous(), None)
        if self.match("NUMBER"):
            literal = self.previous().text
            value = float(literal)
            return LiteralExpr(self.previous(), value)
        if self.match("STRING"):
            literal = self.previous().text
            value = literal[1:-1]
            return LiteralExpr(self.previous(), value)
        if self.match("IDENTIFIER"):
            return VariableExpr(self.previous())
        if self.match("LEFT_PAREN"):
            expr = self.expression()
            self.consume("RIGHT_PAREN", "Expected ')' after expression.")
            return GroupingExpr(expr)
        self.error("Invalid token. Expected an expression.")

    # -------- #
    # Helpers. #
    # -------- #

    def match(self, *token_types):
        for token_type in token_types:
            if self.check(token_type):
                self.advance()
                return True
        return False

    def check(self, token_type):
        if self.is_at_end():
            return False
        return self.peek().type == token_type

    def advance(self):
        if not self.is_at_end():
            self.current += 1
        return self.previous()

    def is_at_end(self):
        return self.peek().type == "EOF"

    def peek(self):
        return self.tokens[self.current]

    def previous(self):
        return self.tokens[self.current - 1]

    def consume(self, token_type, message):
        if self.check(token_type):
            return self.advance()
        self.error(message)

    def error(self, message, token=None):
        token = token or self.peek()
        if token.type == "EOF":
            sys.exit(f"Syntax Error: At EOF. {message}")
        else:
            sys.exit(f"Syntax Error: Line {token.line}, at token '{token.text}'. {message}")



#!/usr/bin/env python3
"""
An experimental programming language.

"""

import os
import io
import re
import setuptools


metapath = os.path.join(os.path.dirname(__file__), 'proto', '__init__.py')
with io.open(metapath, encoding='utf-8') as metafile:
    regex = r'''^__([a-z]+)__ = ["'](.*)["']'''
    meta = dict(re.findall(regex, metafile.read(), flags=re.MULTILINE))


setuptools.setup(
    name = 'protolang',
    version = meta['version'],
    packages = setuptools.find_packages(),
    entry_points = {
        'console_scripts': [
            'proto = proto:main',
        ],
    },
    install_requires = [],
    python_requires = ">=3.8",
    author = 'Darren Mulholland',
    url='https://github.com/dmulholl/proto',
    license = 'Public Domain',
    description = 'An experimental programming language.',
    long_description = __doc__,
    classifiers = [
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'License :: Public Domain',
    ],
)




import sys
from . import builtins


def error(message, token=None):
    if token:
        sys.exit(f"Runtime Error: Line {token.line}. {message}")
    else:
        sys.exit(f"Runtime Error: {message}")


def is_truthy(value):
    return bool(value)


def stringify(value):
    if value is None:
        return "null"
    if value is True:
        return "true"
    if value is False:
        return "false"
    if isinstance(value, float):
        text = str(value)
        if text.endswith('.0'):
            text = text[:-2]
        return text
    return str(value)


def check_operands_are_floats_or_strings(left, right, op_token):
    if isinstance(left, float) and isinstance(right, float):
        return True
    if isinstance(left, str) and isinstance(right, str):
        return True
    error(f"Incompatible operands for the '{op_token.text}' operator.", op_token)


def check_operands_are_floats(left, right, op_token):
    if isinstance(left, float) and isinstance(right, float):
        return True
    error(f"Incompatible operands for the '{op_token.text}' operator.", op_token)


def check_operand_is_float(value, op_token):
    if isinstance(value, float):
        return True
    error(f"Incompatible operand for the '{op_token.text}' operator.", op_token)


class Break(Exception):
    pass


class Continue(Exception):
    pass


class Return(Exception):
    def __init__(self, value):
        self.value = value


class Environment:

    def __init__(self, enclosing_env=None):
        self.values = {}
        self.enclosing_env = enclosing_env

    def define(self, name, value):
        self.values[name] = value

    def get(self, token):
        if token.text in self.values:
            return self.values[token.text]
        elif self.enclosing_env is not None:
            return self.enclosing_env.get(token)
        else:
            error(f"Cannot get undefined variable '{token.text}'.", token)

    def set(self, token, value):
        if token.text in self.values:
            self.values[token.text] = value
        elif self.enclosing_env is not None:
            self.enclosing_env.set(token, value)
        else:
            error(f"Cannot set undefined variable '{token.text}'.", token)

    def get_str(self, name):
        if name in self.values:
            return self.values[name]
        elif self.enclosing_env is not None:
            return self.enclosing_env.get_str(name)
        else:
            error(f"Cannot get undefined variable '{name}'.")


class Interpreter:

    def __init__(self):
        self.env = Environment()
        self.env.define("$clock", builtins.Clock())
        self.env.define("$print", builtins.Print())
        self.env.define("$println", builtins.PrintLine())
        self.env.define("$str", builtins.Stringify())

    def run(self, statements):
        for statement in statements:
            statement.exec(self.env)
        if main := self.env.values.get("$main"):
            if main.arity() != 0:
                error("The $main() function does not accept arguments.")
            main.call([])


class PyroFunction:

    def __init__(self, func_stmt, enclosing_env):
        self.func_stmt = func_stmt
        self.env = enclosing_env

    def __str__(self):
        return f"<PyroFunction {self.func_stmt.name_token.text}>"

    def arity(self):
        return len(self.func_stmt.parameters)

    def call(self, args):
        env = Environment(self.env)
        for i in range(self.arity()):
            env.define(self.func_stmt.parameters[i].text, args[i])
        try:
            self.func_stmt.body.exec(env)
        except Return as ret:
            return ret.value
        return None

    def bind(self, instance_obj):
        instance_env = Environment(self.env)
        instance_env.define("self", instance_obj)
        return PyroFunction(self.func_stmt, instance_env)


class PyroClass:

    def __init__(self, class_stmt, superclass_obj, method_objs, enclosing_env):
        self.class_stmt = class_stmt
        self.superclass_obj = superclass_obj
        self.methods = method_objs
        self.env = enclosing_env
        self.superclass_chain = [self]
        current = superclass_obj
        while current is not None:
            self.superclass_chain.append(current)
            current = current.superclass_obj

    def __str__(self):
        return f"<PyroClass {self.class_stmt.name_token.text}>"

    def call(self, args):
        instance_obj = PyroInstance(self)

        for class_obj in reversed(self.superclass_chain):
            for field_dec in class_obj.class_stmt.field_decs:
                value = None
                if field_dec.initializer is not None:
                    value = field_dec.initializer.eval(class_obj.env)
                instance_obj.fields[field_dec.name_token.text] = value

        constructor = self.find_and_bind_method(instance_obj, "$init")
        if constructor is not None:
            constructor.call(args)

        return instance_obj

    def arity(self):
        constructor = self.find_and_bind_method(None, "$init")
        if constructor is not None:
            return constructor.arity()
        return 0

    def find_and_bind_method(self, instance_obj, method_name):
        if method_name in self.methods:
            return self.methods[method_name].bind(instance_obj)
        if self.superclass_obj is not None:
            return self.superclass_obj.find_and_bind_method(instance_obj, method_name)
        return None


class PyroInstance:

    def __init__(self, class_obj):
       self.class_obj = class_obj
       self.fields = {}

    def __str__(self):
        return f"<PyroInstance {self.class_obj.class_stmt.name_token.text}>"

    def get(self, attr_token):
        if attr_token.text in self.fields:
            return self.fields[attr_token.text]
        bound_func_obj = self.class_obj.find_and_bind_method(self, attr_token.text)
        if bound_func_obj is not None:
            return bound_func_obj
        error(f"Cannot get undefined attribute '{attr_token.text}'.", attr_token)

    def set(self, attr_token, value):
        if attr_token.text in self.fields:
            self.fields[attr_token.text] = value
        else:
            error(f"Cannot set undefined attribute '{attr_token.text}'.", attr_token)




import time
from . import runtime


class Clock:

    def call(self, args):
        return time.perf_counter()

    def arity(self):
        return 0

    def __str__(self):
        return "<ProtoBuiltin $clock>"


class Print:

    def call(self, args):
        print(runtime.stringify(args[0]), end="")

    def arity(self):
        return 1

    def __str__(self):
        return "<ProtoBuiltin $print>"


class PrintLine:

    def call(self, args):
        print(runtime.stringify(args[0]))

    def arity(self):
        return 1

    def __str__(self):
        return "<ProtoBuiltin $println>"


class Stringify:

    def call(self, args):
        return runtime.stringify(args[0])

    def arity(self):
        return 1

    def __str__(self):
        return "<ProtoBuiltin $str>"

